
import java.io.*;
import java.text.SimpleDateFormat;
import java.text.ParseException;
import java.util.*;

/**
 * This program converts TvGuide data files to XMLTV file format (tv_grab_sn).<p>
 * Usage: java ConvDat2XmlTv [file]
 * <pre><font size="3" color="teal" face="serif, monospace">
 * &lt;?xml version="1.0" encoding="ISO-8859-1"?&gt;
 * &lt;!DOCTYPE tv SYSTEM "xmltv.dtd"&gt;
 *
 * &lt;tv source-info-url="http://www.dagenstv.com/" source-data-url="http://www.dagenstv.com/se/frontpage" generator-info-name="ConvDat2XmlTv" generator-info-url="http://hem.bredband.net/tvguide/"&gt;
 *   &lt;channel id="21.dagenstv.com"&gt;
 *     &lt;display-name&gt;SVT1&lt;/display-name&gt;
 *     &lt;icon src="file:///C:/Program Files/TvGuide/images/svt1.gif" /&gt;
 *   &lt;/channel&gt;
 *   &lt;channel id="22.dagenstv.com"&gt;
 *     &lt;display-name&gt;SVT2&lt;/display-name&gt;
 *     &lt;icon src="file:///C:/Program Files/TvGuide/images/svt2.gif" /&gt;
 *   &lt;/channel&gt;
 *   &lt;programme start="20031228233500 +0100" stop="20031229010500 +0100" showview="8289138" channel="21.dagenstv.com"&gt;
 *     &lt;title&gt;Bortom horisonten&lt;/title&gt;
 *     &lt;sub-title&gt;Bis zum Horizont und weiter&lt;/sub-title&gt;
 *     &lt;desc&gt;(Bis zum Horizont und weiter) Tyskt drama frn 1999. Henning Stahnke r en arbetsls fre detta gruvarbetare i det som en gng var sttyskland. Nr hans flickvn dms till fngelse kidnappar han domaren fr att f sin flickvn frislppt. I rollerna: Wolfgang Stumph, Corinna Harfouch, Nina Petri. Regi: Peter Kahane. (Bis zum Horizont und weiter)&lt;/desc&gt;
 *   &lt;/programme&gt;
 *   &lt;programme start="20031228100000 +0100" stop="20031228110000 +0100" showview="27664" channel="22.dagenstv.com"&gt;
 *      &lt;title&gt;Gudstjnst p teckensprk (+29/12)&lt;/title&gt;
 *     &lt;desc&gt;Visuell gudstjnst p dvas villkor frn Johannelunds kyrka i Linkping. Teckensprksprst: Cecilia Nyberg. ven i SVT1 29/12.&lt;/desc&gt;
 *   &lt;/programme&gt;
 * &lt;/tv&gt;
 * </font></pre>
 *
 * @version 1.3
 * @author Henrik Kjell
 *
 * @see <a href="http://hem.bredband.net/tvguide/">http://hem.bredband.net/tvguide/</a><br>
 *      <a href="http://membled.com/work/apps/xmltv/">http://membled.com/work/apps/xmltv/</a>
 */
public class ConvDat2XmlTv {

   static final String SEE_URL_1 = "http://hem.bredband.net/tvguide/";
   static final String SEE_URL_2 = "http://membled.com/work/apps/xmltv/";

   static final String SUPPORTED_VERSION = "1";
   static final String SUBTITLE_VERSION = "1";

   static final int CHANNEL_ID_NDX = 0;
   static final int CHANNEL_NAME_NDX = 1;

   static final String[][] idNameVect = {
      { "1", "3SAT" },
      { "10", "Kanal 5" },
      { "100", "SVT Europa" },
      { "101", "SVT24" },
      { "103", "TV Danmark 1" },
      { "105", "TV Danmark 2" },
      { "106", "Viva" },
      { "107", "Vox" },
      { "108", "WDR" },
      { "11", "Canal+" },
      { "111", "God Channel" },
      { "113", "Kabel 1" },
      { "114", "E!" },
      { "115", "Club" },
      { "117", "Reality TV" },
      { "118", "Discovery Civilisation" },
      { "119", "Discovery Sci-Trek" },
      { "12", "Canal+ Gul" },
      { "120", "Discovery T & A" },
      { "126", "TV2 Zulu" },
      { "13", "Hallmark" },
      { "139", "ORT Int" },
      { "14", "MTV" },
      { "140", "Canal+ Zap" },
      { "141", "Eurosport News" },
      { "142", "TV4 Plus" },
      { "143", "Barnkanalen" },
      { "147", "Disney Channel" },
      { "15", "RTL 2" },
      { "16", "RTL" },
      { "17", "SAT1" },
      { "18", "Sky News" },
      { "2", "BBC Prime" },
      { "21", "SVT1" },
      { "22", "SVT2" },
      { "24", "TCM" },
      { "25", "Travel" },
      { "26", "TV3" },
      { "27", "TV4" },
      { "28", "Viasat Action/Nature" },
      { "29", "TV Finland" },
      { "3", "BBC World" },
      { "31", "YLE TV1" },
      { "33", "TV 1000" },
      { "34", "TV 1000 Cinema" },
      { "36", "YLE TV2" },
      { "37", "TV2 Norge" },
      { "39", "TV3 Finland" },
      { "4", "Cartoon Network" },
      { "40", "TV5" },
      { "41", "TVE" },
      { "42", "ZTV" },
      { "43", "TV8" },
      { "45", "Nickelodeon" },
      { "46", "Mezzo" },
      { "47", "VH-1" },
      { "48", "National Geographic" },
      { "49", "Style" },
      { "5", "CNN" },
      { "51", "ZDF" },
      { "52", "TV Chile" },
      { "53", "ViaSat Sport" },
      { "54", "Animal Planet" },
      { "55", "Canal+ Bl" },
      { "56", "CNBC" },
      { "57", "TV3 Danmark" },
      { "58", "DR1" },
      { "59", "DR2" },
      { "6", "Deutsche Welle" },
      { "60", "Fox Kids" },
      { "62", "TV3 Norge" },
      { "63", "NRK2" },
      { "64", "Rai Uno" },
      { "65", "Rai Due" },
      { "66", "TV Polonia" },
      { "67", "Extreme sports" },
      { "68", "Avante" },
      { "69", "ARD" },
      { "7", "Discovery" },
      { "70", "The Studio" },
      { "71", "TV2 Danmark" },
      { "72", "NRK1" },
      { "73", "TV Norge" },
      { "74", "MTV2" },
      { "75", "ZTV Norge" },
      { "76", "ViaSat Explorer" },
      { "8", "DSF" },
      { "80", "DTU7" },
      { "81", "Bayerisches Fernsehen" },
      { "82", "Fashion TV" },
      { "83", "3+ Danmark" },
      { "84", "Arte" },
      { "85", "Viva Plus" },
      { "86", "France 5" },
      { "87", "The Adult Channel" },
      { "88", "MBC" },
      { "89", "Discovery Mix" },
      { "9", "Eurosport" },
      { "90", "Expo" },
      { "91", "MDR" },
      { "92", "NDR" },
      { "93", "NBC Europe" },
      { "94", "ORB" },
      { "98", "PRO 7" },
      { "99", "Sdwest" },
   };

   Hashtable idTable;
   Hashtable channels;

   String today;
   String tomorrow;

   Hashtable usedChannels;
   StringBuffer buf;

   public ConvDat2XmlTv() {
   }

   private void process(String path)
      throws IOException, ClassNotFoundException {

      // Check if file exists
      File f = new File(path);
      if (!f.exists()) {
         System.out.println("File '" + path + "' not found!");
	 return;
      }

      // Get input stream
      ObjectInputStream in = new ObjectInputStream(new FileInputStream(f));

      // Check data format version
      String ver = (String)in.readObject();
      if (!ver.equals(SUPPORTED_VERSION)) {
	 in.close();
         return;
      }

      // Get "today" and "tomorrow"
      String inpPattern = "yyyy-MM-dd";
      String outPattern = "yyyyMMdd";
      SimpleDateFormat inpSdf = new SimpleDateFormat(inpPattern);
      SimpleDateFormat outSdf = new SimpleDateFormat(outPattern);
      try {
         int ndx = path.lastIndexOf(".") - inpPattern.length();
         today = path.substring(ndx, ndx + inpPattern.length());
         Date date = inpSdf.parse(today);
         GregorianCalendar cal = new GregorianCalendar();
         cal.setTime(date);
         cal.add(Calendar.DATE, 1);
         tomorrow = outSdf.format(cal.getTime());
         today = outSdf.format(date);
      } catch (ParseException e) {
         in.close();
         return;
      }

      // Initiate channel table with default values
      idTable = new Hashtable();
      for (int ndx = 0; ndx < idNameVect.length; ndx++) {
         String[] data = idNameVect[ndx];
         idTable.put(data[CHANNEL_NAME_NDX], data[CHANNEL_ID_NDX]);
      }

      // Get all channels
      channels = new Hashtable();
      int size = in.readInt();
      String lang = null;
      for (int ndx = 0; ndx < size; ndx++) {
	 Channel channel = new Channel();
	 channel.read(in);
         channels.put(channel.name, channel);
         if (channel.id != null) {
            channels.put(channel.id, channel);
         }

         // Update channel id table
         if (channel.id != null) {
            idTable.put(channel.name, channel.id);
            if (lang == null && channel.lang != null) {
               lang = channel.lang;  // language for the first channel
            }
         }
      }
      if (lang == null) {
         lang = "se";
      }

      System.out.println("<?xml version=\"1.0\" encoding=\"ISO-8859-1\"?>");
      System.out.println("<!DOCTYPE tv SYSTEM \"xmltv.dtd\">");
      System.out.println();

      System.out.println("<tv source-info-url=\"http://www.dagenstv.com/\" " +
         "source-data-url=\"http://www.dagenstv.com/" + lang + "/frontpage\" " +
         "generator-info-name=\"ConvDat2XmlTv\" " +
         "generator-info-url=\"" + SEE_URL_1 + "\">");

      // Used channels
      usedChannels = new Hashtable();

      // Programme buffer
      buf = new StringBuffer();

      // Get TableInfo object
      TableInfo ti = new TableInfo();
      ti.read(in);

      in.close();

      // Get 'TV listings read'
      String dateStored = ti.dateStored;
      try {
         Date date = inpSdf.parse(dateStored);
         dateStored = outSdf.format(date);
      } catch (ParseException e) {
         // nada
      }
      System.out.println("  <!-- TV listings read: " + dateStored + " -->");

      // Loop channels with programme information
      Enumeration e = usedChannels.keys();
      while (e.hasMoreElements()) {
         String key = (String)e.nextElement();
         Channel channel = (Channel)usedChannels.get(key);

         String id = channel.id;
         if (id == null) {
            id = (String)idTable.get(channel.name);
            if (id == null) {
               id = "?";
            }
         }
         System.out.println("  <channel id=\"" + id + ".dagenstv.com\">");
         System.out.println("    <display-name>" + channel.name + "</display-name>");
         if (channel.image != null) {
            System.out.println("    <icon src=\"" + channel.image + "\" />");
         }
         System.out.println("  </channel>");
      }

      // Programme
      StringTokenizer st = new StringTokenizer(buf.toString(), "\n", false);
      while (st.hasMoreTokens()) {
         System.out.println(st.nextToken());
      }

      System.out.println("</tv>");
      System.out.println();
   }

/*
   private String decode(String encoded) {

      if (encoded == null) {
	 return null;
      }

      StringBuffer decoded = new StringBuffer();
      for (int i = 0; i < encoded.length(); i++) {
	 char ch = encoded.charAt(i);
	 if (ch == '&' && (i + 1) < encoded.length() &&
             encoded.charAt(i + 1) == '#') {
            int pos = encoded.indexOf(";", i);
            if (pos != -1) {
               int chNo = Integer.valueOf(encoded.substring(i + 2, pos)).intValue();
	       if (chNo == 8212)
		  chNo = (int)'';
	       decoded.append((char)chNo);
               i = pos;
            }
	 } else {
	    decoded.append(ch);
	 }
      }

      String txt = decoded.toString();

      txt = txt.replaceAll("&amp;", "&");
      txt = txt.replaceAll("&lt;", "<");
      txt = txt.replaceAll("&gt;", ">");
      txt = txt.replaceAll("&quot;", "\"");
      txt = txt.replaceAll("&apos;", "\'");
      txt = txt.replaceAll("&acute;", "\'");

      return txt;
   }
*/

   private void appendBuf(String txt) {
      buf.append(txt + "\n");
   }

   /**
    * class TableInfo
    * @version 1.0
    * @author Henrik Kjell
    */
   public class TableInfo {

      private String dateStored;
      private String progInclPath;

      protected TableInfo() {
      }

      protected void read(ObjectInputStream in)
	 throws IOException, ClassNotFoundException {

	 // Data format version
	 String ver = (String)in.readObject();

	 dateStored = (String)in.readObject();

	 if (ver.equals("1"))
	    progInclPath = null;
	 else
	    progInclPath = (String)in.readObject();

	 int size = in.readInt();
	 for (int ndx = 0; ndx < size; ndx++) {
	    Program prg = new Program();
            prg.read(in);
	 }
      }
   }

   /**
    * class Program
    * @version 1.3
    * @author Henrik Kjell
    */
   public class Program {

      protected boolean selected;
      protected int id;
      protected Channel channel;
      protected String start;
      protected String stop;
      protected String name;
      protected String info;
      protected String orgTitle;
      protected String rec;
      protected String showview;
      protected Program more;

      protected void read(ObjectInputStream in)
         throws IOException, ClassNotFoundException {

         // Data format version
         String ver = (String)in.readObject();

         selected = in.readBoolean();

         String channelIdOrName = (String)in.readObject();
         channel = (Channel)channels.get(channelIdOrName);
         if (channel == null) {
            channel = new Channel();
            channel.name = channelIdOrName;
         }

         start = (String)in.readObject();
         stop = (String)in.readObject();
         name = (String)in.readObject();
         info = (String)in.readObject();
         if (!ver.equals(SUBTITLE_VERSION)) {
            orgTitle = (String)in.readObject();
            rec = (String)in.readObject();
         }
         showview = (String)in.readObject();

         if (start.length() > 5) {
            start = start.substring("YYYY-MM-DD HH:MM".length() - 5);
         }
         if (stop != null && stop.length() > 5) {
            stop = stop.substring("YYYY-MM-DD HH:MM".length() - 5);
         }

         String begin = (start.compareTo("05:59") > 0 ? today : tomorrow);
         begin += start.substring(0, 2) + start.substring(3, 5) + "00";

         String end = null;
         if (stop != null && stop.length() != 0) {

            end = (stop.compareTo("05:59") > 0 ? today : tomorrow);
            end += stop.substring(0, 2) + stop.substring(3, 5) + "00";
         }

         String id = channel.id;
         if (id == null) {
            id = (String)idTable.get(channel.name);
            if (id == null) {
               id = "?";
            }
         }

         appendBuf("  <programme start=\"" + begin + " +0100\" " + (end != null ? "stop=\"" + end + " +0100\" " : "") +
            (showview != null ? "showview=\"" + showview + "\" " : "") + "channel=\"" + id + ".dagenstv.com\">");

         if (selected) {
            appendBuf("    <!-- selected=\"" + (selected ? "yes" : "no") + "\" -->");
         }

         if (rec != null) {
//            appendBuf("    <!-- recorded=\"" + rec + "\" -->");
         }

         appendBuf("    <title>" + name + "</title>");

         if (orgTitle != null) {
            appendBuf("    <sub-title>" + orgTitle + "</sub-title>");
         }

         if (info != null) {
            appendBuf("    <desc>" + info + "</desc>");
         }

         // Set channel used
         usedChannels.put(channel.name, channel);

         // Check if more
         String moreChannelId = (String)in.readObject();
         String moreStart = (String)in.readObject();
         if (moreChannelId != null) {
            more = new Program();
            more.name = moreChannelId;
            more.start = moreStart;

            id = (String)idTable.get(more.name);
            if (id == null) {
               id = more.name;
            }
            begin = (more.start.compareTo("05:59") > 0 ? today : tomorrow);
            begin += more.start.substring(0, 2) + more.start.substring(3, 5) + "00 +0100";

            appendBuf("    <!-- more: channel=\"" + id + ".dagenstv.com\"" + " -->");
            appendBuf("    <!-- more: start=\"" + begin + "\"" + " -->");
         }

         appendBuf("  </programme>");
      }

   }

   /**
    * class Channel
    * @version 1.1
    * @author Henrik Kjell
    */
   public class Channel {

      protected String id;
      protected String lang;
      protected String name;
      protected String image;
      protected String info;


      protected Channel() {
      }

      protected void read(ObjectInputStream in)
         throws IOException, ClassNotFoundException {

         // Data format version
         String ver = (String)in.readObject();

         if (!ver.equals("1")) {
            id = (String)in.readObject();
            lang = (String)in.readObject();
         }
         name = (String)in.readObject();
         image = (String)in.readObject();
      }

      public String toString() {
         return name;
      }

   }

   /**
    * Usage: java ConvDat2XmlTv [file]
    */
   public static void main(String[] args) {
      ConvDat2XmlTv dat = new ConvDat2XmlTv();
      try {
         SimpleDateFormat sdf = new SimpleDateFormat("yyyy-MM-dd");
         String today = sdf.format(Calendar.getInstance().getTime());
         String path = null;
         if (args.length >= 1) {
            // Use given file
            path = args[0];
         } else {
            // Use "today" file in default tvguide directory ('Program' or 'Program Files')
            path = File.separator + "Program";
            File f = new File(path);
            if (!f.exists()) {
               path = File.separator + "Program Files";
            }
            path += File.separator + "TvGuide";
            path += File.separator + "download";
            path += File.separator + today + ".dat";
         }
         if (path.indexOf("?") != -1 || path.startsWith("-")) {
            System.out.println("Usage: java ConvDat2XmlTv [file]");
            System.out.println("See Also: " + SEE_URL_1);
            System.out.println("          " + SEE_URL_2);
         } else {
            dat.process(path);
         }
      } catch (IOException e) {
      } catch (ClassNotFoundException e) {
      }
   }

}
